// Do not edit this file! //
// The file was generated from en_US.txt by message-converter.py. //

static char const *
__kmp_i18n_default_meta[] =
    {
        NULL,
        "English",
        "USA",
        "1033",
        "2",
        "20170523",
        NULL
    };

static char const *
__kmp_i18n_default_strings[] =
    {
        NULL,
        "Error",
        "(unknown file)",
        "not a number",
        "bad unit",
        "illegal characters",
        "value too large",
        "value too small",
        "value is not a multiple of 4k",
        "Unknown processor topology",
        "Cannot open /proc/cpuinfo",
        "/proc/cpuinfo",
        "cpuinfo file invalid (No processor records)",
        "cpuinfo file invalid (Too many processor records)",
        "Cannot rewind cpuinfo file",
        "cpuinfo file invalid (long line)",
        "cpuinfo file contains too many entries",
        "cpuinfo file missing processor field",
        "cpuinfo file missing physical id field",
        "cpuinfo file invalid (missing val)",
        "cpuinfo file invalid (duplicate field)",
        "Physical node/pkg/core/thread ids not unique",
        "APIC not present",
        "Invalid cpuid info",
        "APIC ids not unique",
        "Inconsistent cpuid info",
        "Out of heap memory",
        "Memory allocation failed",
        "core",
        "thread",
        "package",
        "node",
        "<undef>",
        "decoding legacy APIC ids",
        "parsing /proc/cpuinfo",
        "value is not defined",
        "Effective settings:",
        "User settings:",
        "warning: pointers or size don't make sense",
        "CPU",
        "TPU",
        "TPUs per package",
        "HT enabled",
        "HT disabled",
        "decoding x2APIC ids",
        "cpuid leaf 11 not supported",
        "cpuid leaf 4 not supported",
        "thread ids not unique",
        "using pthread info",
        "legacy APIC ids not unique",
        "x2APIC ids not unique",
        "OPENMP DISPLAY ENVIRONMENT BEGIN",
        "OPENMP DISPLAY ENVIRONMENT END",
        "[device]",
        "[host]",
        "tile",
        "tiles",
        "threads",
        "cores",
        "socket",
        "sockets",
        "die",
        "dice",
        "module",
        "modules",
        "L1 cache",
        "L1 caches",
        "L2 cache",
        "L2 caches",
        "L3 cache",
        "L3 caches",
        "NUMA domain",
        "NUMA domains",
        "processor group",
        "processor groups",
        "unknown",
        "cpuid leaf 31 not supported",
        "Hwloc api failure",
        "LL cache",
        "LL caches",
        NULL
    };

static char const *
__kmp_i18n_default_formats[] =
    {
        NULL,
        "OMP: Info #%1$d: %2$s\n",
        "OMP: Warning #%1$d: %2$s\n",
        "OMP: Error #%1$d: %2$s\n",
        "OMP: System error #%1$d: %2$s\n",
        "OMP: Hint %1$s\n",
        "%1$s pragma (at %2$s:%3$s():%4$s)",
        NULL
    };

static char const *
__kmp_i18n_default_messages[] =
    {
        NULL,
        "Library is \"serial\".",
        "Cannot open message catalog \"%1$s\":",
        "Default messages will be used.",
        "%1$s: Lock is uninitialized",
        "%1$s: Lock was initialized as simple, but used as nestable",
        "%1$s: Lock was initialized as nestable, but used as simple",
        "%1$s: Lock is already owned by requesting thread",
        "%1$s: Lock is still owned by a thread",
        "%1$s: Attempt to release a lock not owned by any thread",
        "%1$s: Attempt to release a lock owned by another thread",
        "Stack overflow detected for OpenMP thread #%1$d",
        "Stack overlap detected. ",
        "Assertion failure at %1$s(%2$d).",
        "Unable to register a new user thread.",
        "Initializing %1$s, but found %2$s already initialized.",
        "Cannot open file \"%1$s\" for reading:",
        "Getting environment variable \"%1$s\" failed:",
        "Setting environment variable \"%1$s\" failed:",
        "Getting environment failed:",
        "%1$s=\"%2$s\": Wrong value, boolean expected.",
        "No Helper Thread support built in this OMP library.",
        "Helper thread failed to soft terminate.",
        "Buffer overflow detected.",
        "Real-time scheduling policy is not supported.",
        "OMP application is running at maximum priority with real-time scheduling policy. ",
        "Changing priority of the monitor thread failed:",
        "Deadlocks are highly possible due to monitor thread starvation.",
        "Unable to set monitor thread stack size to %1$lu bytes:",
        "Unable to set OMP thread stack size to %1$lu bytes:",
        "Thread attribute initialization failed:",
        "Thread attribute destroying failed:",
        "OMP thread joinable state setting failed:",
        "Monitor thread joinable state setting failed:",
        "System unable to allocate necessary resources for OMP thread:",
        "System unable to allocate necessary resources for the monitor thread:",
        "Unable to terminate OMP thread:",
        "Wrong schedule type %1$d, see <omp.h> or <omp_lib.h> file for the list of values supported.",
        "Unknown scheduling type \"%1$d\".",
        "%1$s value \"%2$s\" is invalid.",
        "%1$s value \"%2$s\" is too small.",
        "%1$s value \"%2$s\" is too large.",
        "%1$s: \"%2$s\" is an invalid value; ignored.",
        "%1$s release value \"%2$s\" is invalid.",
        "%1$s gather value \"%2$s\" is invalid.",
        "%1$s supported only on debug builds; ignored.",
        "Syntax error: Usage: %1$s=[ routine=<func> | filename=<file> | range=<lb>:<ub> | excl_range=<lb>:<ub> ],...",
        "Unbalanced quotes in %1$s.",
        "Empty string specified for %1$s; ignored.",
        "%1$s value is too long; ignored.",
        "%1$s: Invalid clause in \"%2$s\".",
        "Empty clause in %1$s.",
        "%1$s value \"%2$s\" is invalid chunk size.",
        "%1$s value \"%2$s\" is to large chunk size.",
        "%1$s value \"%2$s\" is ignored.",
        "Cannot get processor frequency, using zero KMP_ITT_PREPARE_DELAY.",
        "%1$s must be set prior to first parallel region; ignored.",
        "%1$s: parameter has been specified already, ignoring \"%2$s\".",
        "%1$s: parameter invalid, ignoring \"%2$s\".",
        "%1$s: too many integer parameters specified, ignoring \"%2$s\".",
        "%1$s: too many integer parameters specified for logical or physical type, ignoring \"%2$d\".",
        "%1$s: '%2$s' type does not take any integer parameters, ignoring them.",
        "%1$s: proclist not specified with explicit affinity type, using \"none\".",
        "%1$s: proclist specified, setting affinity type to \"explicit\".",
        "%1$s: proclist specified without \"explicit\" affinity type, proclist ignored.",
        "%1$s: syntax error, not using affinity.",
        "%1$s: range error (zero stride), not using affinity.",
        "%1$s: range error (%2$d > %3$d), not using affinity.",
        "%1$s: range error (%2$d < %3$d & stride < 0), not using affinity.",
        "%1$s: range error ((%2$d-%3$d)/%4$d too big), not using affinity.",
        "%1$s: %2$s is defined. %3$s will be ignored.",
        "%1$s: affinity not supported, using \"disabled\".",
        "%1$s: affinity only supported for Intel(R) Architecture Processors.",
        "%1$s: getaffinity system call not supported.",
        "%1$s: setaffinity system call not supported.",
        "%1$s: pthread_aff_set_np call not found.",
        "%1$s: pthread_get_num_resources_np call not found.",
        "%1$s: the OS kernel does not support affinity.",
        "%1$s: pthread_get_num_resources_np returned %2$d.",
        "%1$s: cannot determine proper affinity mask size.",
        "%1$s=\"%2$s\": %3$s.",
        "%1$s: extra trailing characters ignored: \"%2$s\".",
        "%1$s: unknown method \"%2$s\".",
        "KMP_STATS_TIMER: clock_gettime is undefined, using gettimeofday.",
        "KMP_STATS_TIMER: \"%1$s\" needs additional parameter, e.g. 'clock_gettime,2'. Using gettimeofday.",
        "KMP_STATS_TIMER: clock_gettime parameter \"%1$s\" is invalid, using gettimeofday.",
        "KMP_STATS_TIMER: clock_gettime failed, using gettimeofday.",
        "KMP_STATS_TIMER: clock function unknown (ignoring value \"%1$s\").",
        "Unknown scheduling type detected.",
        "Too many threads to use analytical guided scheduling - switching to iterative guided scheduling.",
        "ittnotify: Lookup of \"%1$s\" function in \"%2$s\" library failed.",
        "ittnotify: Loading \"%1$s\" library failed.",
        "ittnotify: All itt notifications disabled.",
        "ittnotify: Object state itt notifications disabled.",
        "ittnotify: Mark itt notifications disabled.",
        "ittnotify: Unloading \"%1$s\" library failed.",
        "Cannot form a team with %1$d threads, using %2$d instead.",
        "Requested number of active parallel levels \"%1$d\" is negative; ignored.",
        "Requested number of active parallel levels \"%1$d\" exceeds supported limit; the following limit value will be used: \"%1$d\".",
        "kmp_set_library must only be called from the top level serial thread; ignored.",
        "Fatal system error detected.",
        "Out of heap memory.",
        "Clearing __KMP_REGISTERED_LIB env var failed.",
        "Registering library with env var failed.",
        "%1$s value \"%2$d\" will be used.",
        "%1$s value \"%2$u\" will be used.",
        "%1$s value \"%2$s\" will be used.",
        "%1$s value \"%2$s\" will be used.",
        "Mixing other barrier patterns with dist is prohibited. Using dist for all barrier patterns.",
        "%1$s maximum value \"%2$d\" will be used.",
        "%1$s minimum value \"%2$d\" will be used.",
        "Memory allocation failed.",
        "File name too long.",
        "Lock table overflow.",
        "Too many threads to use threadprivate directive.",
        "%1$s: invalid mask.",
        "Wrong definition.",
        "Windows* OS: TLS Set Value failed.",
        "Windows* OS: TLS out of indexes.",
        "PDONE directive must be nested within a DO directive.",
        "Cannot get number of available CPUs.",
        "Assumed number of CPUs is 2.",
        "Error initializing affinity - not using affinity.",
        "Threads may migrate across all available OS procs (granularity setting too coarse).",
        "Ignoring invalid OS proc ID %1$d.",
        "No valid OS proc IDs specified - not using affinity.",
        "%1$s - using \"flat\" OS <-> physical proc mapping.",
        "%1$s: %2$s - using \"flat\" OS <-> physical proc mapping.",
        "%1$s, line %2$d: %3$s - using \"flat\" OS <-> physical proc mapping.",
        "%1$s: %2$s - exiting.",
        "%1$s, line %2$d: %3$s - exiting.",
        "Construct identifier invalid.",
        "Thread identifier invalid.",
        "runtime library not initialized.",
        "Inconsistent THREADPRIVATE common block declarations are non-conforming and are unsupported. Either all threadprivate common blocks must be declared identically, or the largest instance of each threadprivate common block must be referenced first during the run.",
        "Cannot set thread affinity mask.",
        "Cannot set thread priority.",
        "Cannot create thread.",
        "Cannot create event.",
        "Cannot set event.",
        "Cannot close handle.",
        "Unknown library type: %1$d.",
        "Monitor did not reap properly.",
        "Worker thread failed to join.",
        "Cannot change thread affinity mask.",
        "%1$s: Threads may migrate across %2$d innermost levels of machine",
        "%1$s: decrease to %2$d threads",
        "%1$s: increase to %2$d threads",
        "%1$s: Internal thread %2$d bound to OS proc set %3$s",
        "%1$s: Affinity capable, using cpuinfo file",
        "%1$s: Affinity capable, using global cpuid info",
        "%1$s: Affinity capable, using default \"flat\" topology",
        "%1$s: Affinity not capable, using local cpuid info",
        "%1$s: Affinity not capable, using cpuinfo file",
        "%1$s: Affinity not capable, assuming \"flat\" topology",
        "%1$s: Initial OS proc set respected: %2$s",
        "%1$s: Initial OS proc set not respected: %2$s",
        "%1$s: %2$d available OS procs",
        "%1$s: Uniform topology",
        "%1$s: Nonuniform topology",
        "%1$s: %2$d packages x %3$d cores/pkg x %4$d threads/core (%5$d total cores)",
        "%1$s: OS proc to physical thread map ([] => level not in map):",
        "%1$s: OS proc <n> maps to <n>th package core 0",
        "%1$s: OS proc %2$d maps to package %3$d [core %4$d] [thread %5$d]",
        "%1$s: OS proc %2$d maps to [package %3$d] [core %4$d] [thread %5$d]",
        "%1$s: OS proc %2$d maps to [package %3$d] [core %4$d] thread %5$d",
        "%1$s: OS proc %2$d maps to [package %3$d] core %4$d [thread %5$d]",
        "%1$s: OS proc %2$d maps to package %3$d [core %4$d] [thread %5$d]",
        "%1$s: OS proc %2$d maps to [package %3$d] core %4$d thread %5$d",
        "%1$s: OS proc %2$d maps to package %3$d core %4$d [thread %5$d]",
        "%1$s: OS proc %2$d maps to package %3$d [core %4$d] thread %5$d",
        "%1$s: OS proc %2$d maps to package %3$d core %4$d thread %5$d",
        "%1$s: OS proc %2$d maps to %3$s",
        "%1$s: Internal thread %2$d changed affinity mask from %3$s to %4$s",
        "%1$s: OS proc %2$d maps to package %3$d, CPU %4$d, TPU %5$d",
        "%1$s: OS proc %2$d maps to package %3$d, CPU %4$d",
        "%1$s: HT enabled; %2$d packages; %3$d TPU; %4$d TPUs per package",
        "%1$s: HT disabled; %2$d packages",
        "Threads encountered barriers in different order. ",
        "Function %1$s failed:",
        "%1$s: %2$s packages x %3$d cores/pkg x %4$d threads/core (%5$d total cores)",
        "Incompatible message catalog \"%1$s\": Version \"%2$s\" found, version \"%3$s\" expected.",
        "%1$s: ignored because %2$s has been defined",
        "%1$s: overrides %3$s specified before",
        "%1$s: Tiles are only supported if KMP_TOPOLOGY_METHOD=hwloc, using granularity=package instead",
        "%1$s: Tiles requested but were not detected on this HW, using granularity=package instead",
        "%1$s: %2$d packages x %3$d tiles/pkg x %4$d cores/tile x %5$d threads/core (%6$d total cores)",
        "%1$s: %2$d packages x %3$d nodes/pkg x %4$d cores/node x %5$d threads/core (%6$d total cores)",
        "%1$s: %2$d packages x %3$d nodes/pkg x %4$d tiles/node x %5$d cores/tile x %6$d threads/core (%7$d total cores)",
        "OMPT: Cannot determine workshare type; using the default (loop) instead. This issue is fixed in an up-to-date compiler.",
        "Allocator %1$s is not available, will use default allocator.",
        "%1$s: %2$s (%3$d total cores)",
        "%1$s: granularity setting: %2$s does not exist in topology.  Using granularity=%3$s instead.",
        "%1$s: hybrid core type detected: %2$d %3$s cores.",
        "%1$s:   %2$d with core efficiency %3$d.",
        "%1$s must be bound to a work-sharing or work-queuing construct with an \"ordered\" clause",
        "Detected end of %1$s without first executing a corresponding beginning.",
        "Iteration range too large in %1$s.",
        "%1$s must not have a loop increment that evaluates to zero.",
        "Expected end of %1$s; %2$s, however, has most recently begun execution.",
        "%1$s is incorrectly nested within %2$s",
        "%1$s cannot be executed multiple times during execution of one parallel iteration/section of %2$s",
        "%1$s is incorrectly nested within %2$s of the same name",
        "%1$s is incorrectly nested within %2$s that does not have an \"ordered\" clause",
        "%1$s is incorrectly nested within %2$s but not within any of its \"task\" constructs",
        "One thread at %1$s while another thread is at %2$s.",
        "Cannot connect to %1$s",
        "Cannot connect to %1$s - Using %2$s",
        "%1$s does not support %2$s. Continuing without using %2$s.",
        "%1$s does not support %2$s for %3$s. Continuing without using %2$s.",
        "Static %1$s does not support %2$s. Continuing without using %2$s.",
        "KMP_DYNAMIC_MODE=irml cannot be used with KMP_USE_IRML=0",
        "ittnotify: Unknown group \"%2$s\" specified in environment variable \"%1$s\".",
        "ittnotify: Environment variable \"%1$s\" too long: Actual lengths is %2$lu, max allowed length is %3$lu.",
        "%1$s: Affinity capable, using global cpuid leaf 11 info",
        "%1$s: Affinity not capable, using local cpuid leaf 11 info",
        "%1$s: %2$s.",
        "%1$s: %2$s - %3$s.",
        "%1$s: OS proc to physical thread map:",
        "%1$s: using \"flat\" OS <-> physical proc mapping.",
        "%1$s: parsing %2$s.",
        "%1$s - exiting.",
        "Incompatible %1$s library with version %2$s found.",
        "ittnotify: Function %1$s failed:",
        "ittnotify: Error #%1$d.",
        "%1$s must be set prior to first parallel region or certain API calls; ignored.",
        "Lock initialized at %1$s(%2$d) was not destroyed",
        "Cannot determine machine load balance - Using %1$s",
        "%1$s: Affinity not capable, using pthread info",
        "%1$s: Affinity capable, using pthread info",
        "Loading \"%1$s\" library failed:",
        "Lookup of \"%1$s\" function failed:",
        "Buffer too small.",
        "Error #%1$d.",
        "%1$s: Invalid symbols found. Check the value \"%2$s\".",
        "%1$s: Spaces between digits are not allowed \"%2$s\".",
        "%1$s: %2$s - parsing %3$s.",
        "%1$s cannot be specified via kmp_set_defaults() on this machine because it has more than one processor group.",
        "Cannot use affinity type \"%1$s\" with multiple Windows* OS processor groups, using \"%2$s\".",
        "Cannot use affinity granularity \"%1$s\" with multiple Windows* OS processor groups, using \"%2$s\".",
        "%1$s: Mapping Windows* OS processor group <i> proc <j> to OS proc 64*<i>+<j>.",
        "%1$s: OS proc %2$d maps to Windows* OS processor group %3$d proc %4$d",
        "%1$s: Affinity balanced is not available.",
        "%1$s: granularity=core will be used.",
        "%1$s must be set prior to first OMP lock call or critical section; ignored.",
        "futex system call not supported; %1$s=%2$s ignored.",
        "%1$s: granularity=%2$s will be used.",
        "%1$s: invalid value \"%2$s\", valid format is \"N<item>[@N][,...][,Nt] (<item> can be S, N, L2, C, T  for Socket, NUMA Node, L2 Cache, Core, Thread)\".",
        "KMP_HW_SUBSET ignored: unsupported architecture.",
        "KMP_HW_SUBSET ignored: too many cores requested.",
        "%1$s: syntax error, using %2$s.",
        "%1$s: Adaptive locks are not supported; using queuing.",
        "%1$s: Invalid symbols found. Check the value \"%2$s\".",
        "%1$s: Spaces between digits are not allowed \"%2$s\".",
        "%1$s: pid %2$d tid %3$d thread %4$d bound to OS proc set %5$s",
        "%1$s error: parallel loop increment and condition are inconsistent.",
        "libgomp cancellation is not currently supported.",
        "KMP_HW_SUBSET ignored: non-uniform topology.",
        "KMP_HW_SUBSET ignored: only three-level topology is supported.",
        "%1$s: granularity=%2$s is not supported with KMP_TOPOLOGY_METHOD=group. Using \"granularity=fine\".",
        "%1$s: granularity=group is not supported with KMP_AFFINITY=%2$s. Using \"granularity=core\".",
        "KMP_HW_SUBSET ignored: too many sockets requested.",
        "KMP_HW_SUBSET \"o\" offset designator deprecated, please use @ prefix for offset value.",
        "%1$s: Affinity capable, using hwloc.",
        "%1$s: Ignoring hwloc mechanism.",
        "%1$s: Hwloc failed in %2$s. Relying on internal affinity mechanisms.",
        "%1$s must be set prior to OpenMP runtime library initialization; ignored.",
        "You have enabled the use of umonitor/umwait. If the CPU doesn't have that enabled you'll get an illegal instruction exception.",
        "%1$s variable deprecated, please use %2$s instead.",
        "KMP_FORCE_REDUCTION: %1$s method is not supported; using critical.",
        "KMP_HW_SUBSET ignored: unsupported item requested for non-HWLOC topology method (KMP_TOPOLOGY_METHOD)",
        "KMP_HW_SUBSET ignored: too many NUMA Nodes requested.",
        "KMP_HW_SUBSET ignored: too many L2 Caches requested.",
        "KMP_HW_SUBSET ignored: too many Procs requested.",
        "Hierarchy ignored: unsupported level: %1$s.",
        "OMP: pid %1$s tid %2$s thread %3$s bound to OS proc set {%4$s}",
        "%1$s routine deprecated, please use %2$s instead.",
        "libgomp compatibility layer does not support OpenMP feature: %1$s",
        "KMP_HW_SUBSET ignored: too many Dies requested.",
        "%1$s: Affinity capable, using global cpuid leaf %2$d info",
        "%1$s: Affinity not capable, using local cpuid leaf %2$d info",
        "%1$s: Affinity not capable, using hwloc.",
        "%1$s: Encountered user-directed error: %2$s.",
        "%1$s: Encountered user-directed warning: %2$s.",
        "Failed to create teams between lower bound (%1$d) and upper bound (%2$d).",
        "KMP_HW_SUBSET ignored: %1$s: too many requested.",
        "KMP_HW_SUBSET ignored: %1$s: level not detected in machine topology.",
        "KMP_HW_SUBSET ignored: %1$s, %2$s: layers are equivalent, please only specify one.",
        "KMP_HW_SUBSET ignored: %1$s layer should come after %2$s.",
        "%1$s: topology layer \"%2$s\" is equivalent to \"%3$s\".",
        "%1$s: granularity=%2$s is too coarse, setting granularity=group.",
        "%1$s: \"%2$s\" value is deprecated. Please use \"%3$s\" instead.",
        "num_teams value must be positive, it is %1$d, using %2$d instead.",
        "KMP_HW_SUBSET ignored: %1$s, %2$s: attributes are ambiguous, please only specify one.",
        "KMP_HW_SUBSET ignored: %1$s: attribute specified more than once.",
        "KMP_HW_SUBSET ignored: %1$s: attribute value %2$s is invalid.",
        "KMP_HW_SUBSET ignored: all hardware resources would be filtered, please reduce the filter.",
        "KMP_HW_SUBSET ignored: Too many attributes specified. This machine is not a hybrid architecutre.",
        "KMP_HW_SUBSET: ignoring %1$s attribute. This machine is not a hybrid architecutre.",
        "Target memory not available, will use default allocator.",
        "%1$s ignored: This machine is not a hybrid architecutre. Using \"%2$s\" instead.",
        "%1$s ignored: %2$s is not available. Using \"%3$s\" instead.",
        NULL
    };

static char const *
__kmp_i18n_default_hints[] =
    {
        NULL,
        "Please submit a bug report with this message, compile and run commands used, and machine configuration info including native compiler and operating system versions. Faster response will be obtained by including all program sources. For information on submitting this issue, please see https://github.com/llvm/llvm-project/issues/.",
        "Check NLSPATH environment variable, its value is \"%1$s\".",
        "Please try changing the shell stack limit or adjusting the OMP_STACKSIZE environment variable.",
        "Consider unsetting KMP_DEVICE_THREAD_LIMIT (KMP_ALL_THREADS), KMP_TEAMS_THREAD_LIMIT, and OMP_THREAD_LIMIT (if any are set).",
        "Consider setting KMP_ALL_THREADPRIVATE to a value larger than %1$d.",
        "This could also be due to a system-related limit on the number of threads.",
        "This means that multiple copies of the OpenMP runtime have been linked into the program. That is dangerous, since it can degrade performance or cause incorrect results. The best thing to do is to ensure that only a single OpenMP runtime is linked into the process, e.g. by avoiding static linking of the OpenMP runtime in any library. As an unsafe, unsupported, undocumented workaround you can set the environment variable KMP_DUPLICATE_LIB_OK=TRUE to allow the program to continue to execute, but that may cause crashes or silently produce incorrect results. For more information, please see http://openmp.llvm.org/",
        "This name is specified in environment variable KMP_CPUINFO_FILE.",
        "Seems application required too much memory.",
        "Use \"0\", \"FALSE\". \".F.\", \"off\", \"no\" as false values, \"1\", \"TRUE\", \".T.\", \"on\", \"yes\" as true values.",
        "Perhaps too many threads.",
        "Decrease priority of application. This will allow the monitor thread run at higher priority than other threads.",
        "Try changing KMP_MONITOR_STACKSIZE or the shell stack limit.",
        "Try changing OMP_STACKSIZE and/or the shell stack limit.",
        "Try increasing OMP_STACKSIZE or the shell stack limit.",
        "Try decreasing OMP_STACKSIZE.",
        "Try decreasing the value of OMP_NUM_THREADS.",
        "Try increasing KMP_MONITOR_STACKSIZE.",
        "Try decreasing KMP_MONITOR_STACKSIZE.",
        "Try decreasing the number of threads in use simultaneously.",
        "Will use default schedule type (%1$s).",
        "It could be a result of using an older OMP library with a newer compiler or memory corruption. You may check the proper OMP library is linked to the application.",
        "Check %1$s environment variable, its value is \"%2$s\".",
        "You may want to use an %1$s library that supports %2$s interface with version %3$s.",
        "You may want to use an %1$s library with version %2$s.",
        "System error #193 is \"Bad format of EXE or DLL file\". Usually it means the file is found, but it is corrupted or a file for another architecture. Check whether \"%1$s\" is a file for %2$s architecture.",
        "System-related limit on the number of threads.",
        "Try setting new bounds (preferably less than or equal to %1$d) for num_teams clause.",
        "Valid values are from %1$d to %2$d.",
        NULL
    };

struct kmp_i18n_section {
    int           size;
    char const ** str;
}; // struct kmp_i18n_section
typedef struct kmp_i18n_section  kmp_i18n_section_t;

static kmp_i18n_section_t
__kmp_i18n_sections[] =
    {
        { 0, NULL },
        { 5, __kmp_i18n_default_meta },
        { 79, __kmp_i18n_default_strings },
        { 6, __kmp_i18n_default_formats },
        { 301, __kmp_i18n_default_messages },
        { 29, __kmp_i18n_default_hints },
        { 0, NULL }
    };

struct kmp_i18n_table {
    int                   size;
    kmp_i18n_section_t *  sect;
}; // struct kmp_i18n_table
typedef struct kmp_i18n_table  kmp_i18n_table_t;

static kmp_i18n_table_t __kmp_i18n_default_table =
    {
        5,
        __kmp_i18n_sections
    };

// end of file //
