// Do not edit this file! //
// The file was generated from en_US.txt by message-converter.py. //

enum kmp_i18n_id {

    // A special id for absence of message.
    kmp_i18n_null = 0,

    // Set #1, meta.
    kmp_i18n_prp_first = 65536,
    kmp_i18n_prp_Language,
    kmp_i18n_prp_Country,
    kmp_i18n_prp_LangId,
    kmp_i18n_prp_Version,
    kmp_i18n_prp_Revision,
    kmp_i18n_prp_last,

    // Set #2, strings.
    kmp_i18n_str_first = 131072,
    kmp_i18n_str_Error,
    kmp_i18n_str_UnknownFile,
    kmp_i18n_str_NotANumber,
    kmp_i18n_str_BadUnit,
    kmp_i18n_str_IllegalCharacters,
    kmp_i18n_str_ValueTooLarge,
    kmp_i18n_str_ValueTooSmall,
    kmp_i18n_str_NotMultiple4K,
    kmp_i18n_str_UnknownTopology,
    kmp_i18n_str_CantOpenCpuinfo,
    kmp_i18n_str_ProcCpuinfo,
    kmp_i18n_str_NoProcRecords,
    kmp_i18n_str_TooManyProcRecords,
    kmp_i18n_str_CantRewindCpuinfo,
    kmp_i18n_str_LongLineCpuinfo,
    kmp_i18n_str_TooManyEntries,
    kmp_i18n_str_MissingProcField,
    kmp_i18n_str_MissingPhysicalIDField,
    kmp_i18n_str_MissingValCpuinfo,
    kmp_i18n_str_DuplicateFieldCpuinfo,
    kmp_i18n_str_PhysicalIDsNotUnique,
    kmp_i18n_str_ApicNotPresent,
    kmp_i18n_str_InvalidCpuidInfo,
    kmp_i18n_str_OBSOLETE1,
    kmp_i18n_str_InconsistentCpuidInfo,
    kmp_i18n_str_OutOfHeapMemory,
    kmp_i18n_str_MemoryAllocFailed,
    kmp_i18n_str_Core,
    kmp_i18n_str_Thread,
    kmp_i18n_str_Package,
    kmp_i18n_str_Node,
    kmp_i18n_str_OBSOLETE2,
    kmp_i18n_str_DecodingLegacyAPIC,
    kmp_i18n_str_OBSOLETE3,
    kmp_i18n_str_NotDefined,
    kmp_i18n_str_EffectiveSettings,
    kmp_i18n_str_UserSettings,
    kmp_i18n_str_StorageMapWarning,
    kmp_i18n_str_OBSOLETE4,
    kmp_i18n_str_OBSOLETE5,
    kmp_i18n_str_OBSOLETE6,
    kmp_i18n_str_OBSOLETE7,
    kmp_i18n_str_OBSOLETE8,
    kmp_i18n_str_Decodingx2APIC,
    kmp_i18n_str_NoLeaf11Support,
    kmp_i18n_str_NoLeaf4Support,
    kmp_i18n_str_ThreadIDsNotUnique,
    kmp_i18n_str_UsingPthread,
    kmp_i18n_str_LegacyApicIDsNotUnique,
    kmp_i18n_str_x2ApicIDsNotUnique,
    kmp_i18n_str_DisplayEnvBegin,
    kmp_i18n_str_DisplayEnvEnd,
    kmp_i18n_str_Device,
    kmp_i18n_str_Host,
    kmp_i18n_str_Tile,
    kmp_i18n_str_Tiles,
    kmp_i18n_str_Threads,
    kmp_i18n_str_Cores,
    kmp_i18n_str_Socket,
    kmp_i18n_str_Sockets,
    kmp_i18n_str_Die,
    kmp_i18n_str_Dice,
    kmp_i18n_str_Module,
    kmp_i18n_str_Modules,
    kmp_i18n_str_L1Cache,
    kmp_i18n_str_L1Caches,
    kmp_i18n_str_L2Cache,
    kmp_i18n_str_L2Caches,
    kmp_i18n_str_L3Cache,
    kmp_i18n_str_L3Caches,
    kmp_i18n_str_NumaDomain,
    kmp_i18n_str_NumaDomains,
    kmp_i18n_str_ProcGroup,
    kmp_i18n_str_ProcGroups,
    kmp_i18n_str_Unknown,
    kmp_i18n_str_NoLeaf31Support,
    kmp_i18n_str_HwlocFailed,
    kmp_i18n_str_LLCache,
    kmp_i18n_str_LLCaches,
    kmp_i18n_str_last,

    // Set #3, formats.
    kmp_i18n_fmt_first = 196608,
    kmp_i18n_fmt_Info,
    kmp_i18n_fmt_Warning,
    kmp_i18n_fmt_Fatal,
    kmp_i18n_fmt_SysErr,
    kmp_i18n_fmt_Hint,
    kmp_i18n_fmt_Pragma,
    kmp_i18n_fmt_last,

    // Set #4, messages.
    kmp_i18n_msg_first = 262144,
    kmp_i18n_msg_LibraryIsSerial,
    kmp_i18n_msg_CantOpenMessageCatalog,
    kmp_i18n_msg_WillUseDefaultMessages,
    kmp_i18n_msg_LockIsUninitialized,
    kmp_i18n_msg_LockSimpleUsedAsNestable,
    kmp_i18n_msg_LockNestableUsedAsSimple,
    kmp_i18n_msg_LockIsAlreadyOwned,
    kmp_i18n_msg_LockStillOwned,
    kmp_i18n_msg_LockUnsettingFree,
    kmp_i18n_msg_LockUnsettingSetByAnother,
    kmp_i18n_msg_StackOverflow,
    kmp_i18n_msg_StackOverlap,
    kmp_i18n_msg_AssertionFailure,
    kmp_i18n_msg_CantRegisterNewThread,
    kmp_i18n_msg_DuplicateLibrary,
    kmp_i18n_msg_CantOpenFileForReading,
    kmp_i18n_msg_CantGetEnvVar,
    kmp_i18n_msg_CantSetEnvVar,
    kmp_i18n_msg_CantGetEnvironment,
    kmp_i18n_msg_BadBoolValue,
    kmp_i18n_msg_SSPNotBuiltIn,
    kmp_i18n_msg_SPPSotfTerminateFailed,
    kmp_i18n_msg_BufferOverflow,
    kmp_i18n_msg_RealTimeSchedNotSupported,
    kmp_i18n_msg_RunningAtMaxPriority,
    kmp_i18n_msg_CantChangeMonitorPriority,
    kmp_i18n_msg_MonitorWillStarve,
    kmp_i18n_msg_CantSetMonitorStackSize,
    kmp_i18n_msg_CantSetWorkerStackSize,
    kmp_i18n_msg_CantInitThreadAttrs,
    kmp_i18n_msg_CantDestroyThreadAttrs,
    kmp_i18n_msg_CantSetWorkerState,
    kmp_i18n_msg_CantSetMonitorState,
    kmp_i18n_msg_NoResourcesForWorkerThread,
    kmp_i18n_msg_NoResourcesForMonitorThread,
    kmp_i18n_msg_CantTerminateWorkerThread,
    kmp_i18n_msg_ScheduleKindOutOfRange,
    kmp_i18n_msg_UnknownSchedulingType,
    kmp_i18n_msg_InvalidValue,
    kmp_i18n_msg_SmallValue,
    kmp_i18n_msg_LargeValue,
    kmp_i18n_msg_StgInvalidValue,
    kmp_i18n_msg_BarrReleaseValueInvalid,
    kmp_i18n_msg_BarrGatherValueInvalid,
    kmp_i18n_msg_OBSOLETE9,
    kmp_i18n_msg_ParRangeSyntax,
    kmp_i18n_msg_UnbalancedQuotes,
    kmp_i18n_msg_EmptyString,
    kmp_i18n_msg_LongValue,
    kmp_i18n_msg_InvalidClause,
    kmp_i18n_msg_EmptyClause,
    kmp_i18n_msg_InvalidChunk,
    kmp_i18n_msg_LargeChunk,
    kmp_i18n_msg_IgnoreChunk,
    kmp_i18n_msg_CantGetProcFreq,
    kmp_i18n_msg_EnvParallelWarn,
    kmp_i18n_msg_AffParamDefined,
    kmp_i18n_msg_AffInvalidParam,
    kmp_i18n_msg_AffManyParams,
    kmp_i18n_msg_AffManyParamsForLogic,
    kmp_i18n_msg_AffNoParam,
    kmp_i18n_msg_AffNoProcList,
    kmp_i18n_msg_AffProcListNoType,
    kmp_i18n_msg_AffProcListNotExplicit,
    kmp_i18n_msg_AffSyntaxError,
    kmp_i18n_msg_AffZeroStride,
    kmp_i18n_msg_AffStartGreaterEnd,
    kmp_i18n_msg_AffStrideLessZero,
    kmp_i18n_msg_AffRangeTooBig,
    kmp_i18n_msg_OBSOLETE10,
    kmp_i18n_msg_AffNotSupported,
    kmp_i18n_msg_OBSOLETE11,
    kmp_i18n_msg_GetAffSysCallNotSupported,
    kmp_i18n_msg_SetAffSysCallNotSupported,
    kmp_i18n_msg_OBSOLETE12,
    kmp_i18n_msg_OBSOLETE13,
    kmp_i18n_msg_OBSOLETE14,
    kmp_i18n_msg_OBSOLETE15,
    kmp_i18n_msg_AffCantGetMaskSize,
    kmp_i18n_msg_ParseSizeIntWarn,
    kmp_i18n_msg_ParseExtraCharsWarn,
    kmp_i18n_msg_UnknownForceReduction,
    kmp_i18n_msg_TimerUseGettimeofday,
    kmp_i18n_msg_TimerNeedMoreParam,
    kmp_i18n_msg_TimerInvalidParam,
    kmp_i18n_msg_TimerGettimeFailed,
    kmp_i18n_msg_TimerUnknownFunction,
    kmp_i18n_msg_UnknownSchedTypeDetected,
    kmp_i18n_msg_DispatchManyThreads,
    kmp_i18n_msg_IttLookupFailed,
    kmp_i18n_msg_IttLoadLibFailed,
    kmp_i18n_msg_IttAllNotifDisabled,
    kmp_i18n_msg_IttObjNotifDisabled,
    kmp_i18n_msg_IttMarkNotifDisabled,
    kmp_i18n_msg_IttUnloadLibFailed,
    kmp_i18n_msg_CantFormThrTeam,
    kmp_i18n_msg_ActiveLevelsNegative,
    kmp_i18n_msg_ActiveLevelsExceedLimit,
    kmp_i18n_msg_SetLibraryIncorrectCall,
    kmp_i18n_msg_FatalSysError,
    kmp_i18n_msg_OutOfHeapMemory,
    kmp_i18n_msg_OBSOLETE16,
    kmp_i18n_msg_OBSOLETE17,
    kmp_i18n_msg_Using_int_Value,
    kmp_i18n_msg_Using_uint_Value,
    kmp_i18n_msg_Using_uint64_Value,
    kmp_i18n_msg_Using_str_Value,
    kmp_i18n_msg_BarrierPatternOverride,
    kmp_i18n_msg_MaxValueUsing,
    kmp_i18n_msg_MinValueUsing,
    kmp_i18n_msg_MemoryAllocFailed,
    kmp_i18n_msg_FileNameTooLong,
    kmp_i18n_msg_OBSOLETE18,
    kmp_i18n_msg_ManyThreadsForTPDirective,
    kmp_i18n_msg_AffinityInvalidMask,
    kmp_i18n_msg_WrongDefinition,
    kmp_i18n_msg_TLSSetValueFailed,
    kmp_i18n_msg_TLSOutOfIndexes,
    kmp_i18n_msg_OBSOLETE19,
    kmp_i18n_msg_CantGetNumAvailCPU,
    kmp_i18n_msg_AssumedNumCPU,
    kmp_i18n_msg_ErrorInitializeAffinity,
    kmp_i18n_msg_AffThreadsMayMigrate,
    kmp_i18n_msg_AffIgnoreInvalidProcID,
    kmp_i18n_msg_AffNoValidProcID,
    kmp_i18n_msg_UsingFlatOS,
    kmp_i18n_msg_UsingFlatOSFile,
    kmp_i18n_msg_UsingFlatOSFileLine,
    kmp_i18n_msg_FileMsgExiting,
    kmp_i18n_msg_FileLineMsgExiting,
    kmp_i18n_msg_ConstructIdentInvalid,
    kmp_i18n_msg_ThreadIdentInvalid,
    kmp_i18n_msg_RTLNotInitialized,
    kmp_i18n_msg_TPCommonBlocksInconsist,
    kmp_i18n_msg_CantSetThreadAffMask,
    kmp_i18n_msg_CantSetThreadPriority,
    kmp_i18n_msg_CantCreateThread,
    kmp_i18n_msg_CantCreateEvent,
    kmp_i18n_msg_CantSetEvent,
    kmp_i18n_msg_CantCloseHandle,
    kmp_i18n_msg_UnknownLibraryType,
    kmp_i18n_msg_ReapMonitorError,
    kmp_i18n_msg_ReapWorkerError,
    kmp_i18n_msg_ChangeThreadAffMaskError,
    kmp_i18n_msg_ThreadsMigrate,
    kmp_i18n_msg_DecreaseToThreads,
    kmp_i18n_msg_IncreaseToThreads,
    kmp_i18n_msg_OBSOLETE20,
    kmp_i18n_msg_AffCapableUseCpuinfo,
    kmp_i18n_msg_AffUseGlobCpuid,
    kmp_i18n_msg_AffCapableUseFlat,
    kmp_i18n_msg_AffNotCapableUseLocCpuid,
    kmp_i18n_msg_AffNotCapableUseCpuinfo,
    kmp_i18n_msg_AffFlatTopology,
    kmp_i18n_msg_InitOSProcSetRespect,
    kmp_i18n_msg_InitOSProcSetNotRespect,
    kmp_i18n_msg_AvailableOSProc,
    kmp_i18n_msg_Uniform,
    kmp_i18n_msg_NonUniform,
    kmp_i18n_msg_Topology,
    kmp_i18n_msg_OBSOLETE21,
    kmp_i18n_msg_OSProcToPackage,
    kmp_i18n_msg_OBSOLETE22,
    kmp_i18n_msg_OBSOLETE23,
    kmp_i18n_msg_OBSOLETE24,
    kmp_i18n_msg_OBSOLETE25,
    kmp_i18n_msg_OBSOLETE26,
    kmp_i18n_msg_OBSOLETE27,
    kmp_i18n_msg_OBSOLETE28,
    kmp_i18n_msg_OBSOLETE29,
    kmp_i18n_msg_OBSOLETE30,
    kmp_i18n_msg_OSProcMapToPack,
    kmp_i18n_msg_OBSOLETE31,
    kmp_i18n_msg_OBSOLETE32,
    kmp_i18n_msg_OBSOLETE33,
    kmp_i18n_msg_OBSOLETE34,
    kmp_i18n_msg_OBSOLETE35,
    kmp_i18n_msg_BarriersInDifferentOrder,
    kmp_i18n_msg_FunctionError,
    kmp_i18n_msg_TopologyExtra,
    kmp_i18n_msg_WrongMessageCatalog,
    kmp_i18n_msg_StgIgnored,
    kmp_i18n_msg_OBSOLETE36,
    kmp_i18n_msg_AffTilesNoHWLOC,
    kmp_i18n_msg_AffTilesNoTiles,
    kmp_i18n_msg_TopologyExtraTile,
    kmp_i18n_msg_TopologyExtraNode,
    kmp_i18n_msg_TopologyExtraNoTi,
    kmp_i18n_msg_OmptOutdatedWorkshare,
    kmp_i18n_msg_OmpNoAllocator,
    kmp_i18n_msg_TopologyGeneric,
    kmp_i18n_msg_AffGranularityBad,
    kmp_i18n_msg_TopologyHybrid,
    kmp_i18n_msg_TopologyHybridCoreEff,
    kmp_i18n_msg_CnsBoundToWorksharing,
    kmp_i18n_msg_CnsDetectedEnd,
    kmp_i18n_msg_CnsIterationRangeTooLarge,
    kmp_i18n_msg_CnsLoopIncrZeroProhibited,
    kmp_i18n_msg_CnsExpectedEnd,
    kmp_i18n_msg_CnsInvalidNesting,
    kmp_i18n_msg_CnsMultipleNesting,
    kmp_i18n_msg_CnsNestingSameName,
    kmp_i18n_msg_CnsNoOrderedClause,
    kmp_i18n_msg_CnsNotInTaskConstruct,
    kmp_i18n_msg_CnsThreadsAtBarrier,
    kmp_i18n_msg_CantConnect,
    kmp_i18n_msg_CantConnectUsing,
    kmp_i18n_msg_LibNotSupport,
    kmp_i18n_msg_LibNotSupportFor,
    kmp_i18n_msg_StaticLibNotSupport,
    kmp_i18n_msg_OBSOLETE37,
    kmp_i18n_msg_IttUnknownGroup,
    kmp_i18n_msg_IttEnvVarTooLong,
    kmp_i18n_msg_OBSOLETE38,
    kmp_i18n_msg_OBSOLETE39,
    kmp_i18n_msg_AffInfoStr,
    kmp_i18n_msg_AffInfoStrStr,
    kmp_i18n_msg_OSProcToPhysicalThreadMap,
    kmp_i18n_msg_AffUsingFlatOS,
    kmp_i18n_msg_AffParseFilename,
    kmp_i18n_msg_MsgExiting,
    kmp_i18n_msg_IncompatibleLibrary,
    kmp_i18n_msg_IttFunctionError,
    kmp_i18n_msg_IttUnknownError,
    kmp_i18n_msg_EnvMiddleWarn,
    kmp_i18n_msg_CnsLockNotDestroyed,
    kmp_i18n_msg_CantLoadBalUsing,
    kmp_i18n_msg_AffNotCapableUsePthread,
    kmp_i18n_msg_AffUsePthread,
    kmp_i18n_msg_OBSOLETE40,
    kmp_i18n_msg_OBSOLETE41,
    kmp_i18n_msg_OBSOLETE42,
    kmp_i18n_msg_OBSOLETE43,
    kmp_i18n_msg_NthSyntaxError,
    kmp_i18n_msg_NthSpacesNotAllowed,
    kmp_i18n_msg_AffStrParseFilename,
    kmp_i18n_msg_OBSOLETE44,
    kmp_i18n_msg_AffTypeCantUseMultGroups,
    kmp_i18n_msg_AffGranCantUseMultGroups,
    kmp_i18n_msg_AffWindowsProcGroupMap,
    kmp_i18n_msg_AffOSProcToGroup,
    kmp_i18n_msg_AffBalancedNotAvail,
    kmp_i18n_msg_OBSOLETE45,
    kmp_i18n_msg_EnvLockWarn,
    kmp_i18n_msg_FutexNotSupported,
    kmp_i18n_msg_AffGranUsing,
    kmp_i18n_msg_AffHWSubsetInvalid,
    kmp_i18n_msg_AffHWSubsetUnsupported,
    kmp_i18n_msg_AffHWSubsetManyCores,
    kmp_i18n_msg_SyntaxErrorUsing,
    kmp_i18n_msg_AdaptiveNotSupported,
    kmp_i18n_msg_EnvSyntaxError,
    kmp_i18n_msg_EnvSpacesNotAllowed,
    kmp_i18n_msg_BoundToOSProcSet,
    kmp_i18n_msg_CnsLoopIncrIllegal,
    kmp_i18n_msg_NoGompCancellation,
    kmp_i18n_msg_AffHWSubsetNonUniform,
    kmp_i18n_msg_AffHWSubsetNonThreeLevel,
    kmp_i18n_msg_AffGranTopGroup,
    kmp_i18n_msg_AffGranGroupType,
    kmp_i18n_msg_AffHWSubsetManySockets,
    kmp_i18n_msg_AffHWSubsetDeprecated,
    kmp_i18n_msg_AffUsingHwloc,
    kmp_i18n_msg_AffIgnoringHwloc,
    kmp_i18n_msg_AffHwlocErrorOccurred,
    kmp_i18n_msg_EnvSerialWarn,
    kmp_i18n_msg_EnvMwaitWarn,
    kmp_i18n_msg_EnvVarDeprecated,
    kmp_i18n_msg_RedMethodNotSupported,
    kmp_i18n_msg_AffHWSubsetNoHWLOC,
    kmp_i18n_msg_AffHWSubsetManyNodes,
    kmp_i18n_msg_AffHWSubsetManyTiles,
    kmp_i18n_msg_AffHWSubsetManyProcs,
    kmp_i18n_msg_HierSchedInvalid,
    kmp_i18n_msg_AffFormatDefault,
    kmp_i18n_msg_APIDeprecated,
    kmp_i18n_msg_GompFeatureNotSupported,
    kmp_i18n_msg_AffHWSubsetManyDies,
    kmp_i18n_msg_AffUseGlobCpuidL,
    kmp_i18n_msg_AffNotCapableUseLocCpuidL,
    kmp_i18n_msg_AffNotUsingHwloc,
    kmp_i18n_msg_UserDirectedError,
    kmp_i18n_msg_UserDirectedWarning,
    kmp_i18n_msg_FailedToCreateTeam,
    kmp_i18n_msg_AffHWSubsetManyGeneric,
    kmp_i18n_msg_AffHWSubsetNotExistGeneric,
    kmp_i18n_msg_AffHWSubsetEqvLayers,
    kmp_i18n_msg_AffHWSubsetOutOfOrder,
    kmp_i18n_msg_AffEqualTopologyTypes,
    kmp_i18n_msg_AffGranTooCoarseProcGroup,
    kmp_i18n_msg_StgDeprecatedValue,
    kmp_i18n_msg_NumTeamsNotPositive,
    kmp_i18n_msg_AffHWSubsetIncompat,
    kmp_i18n_msg_AffHWSubsetAttrRepeat,
    kmp_i18n_msg_AffHWSubsetAttrInvalid,
    kmp_i18n_msg_AffHWSubsetAllFiltered,
    kmp_i18n_msg_AffHWSubsetAttrsNonHybrid,
    kmp_i18n_msg_AffHWSubsetIgnoringAttr,
    kmp_i18n_msg_TargetMemNotAvailable,
    kmp_i18n_msg_AffIgnoringNonHybrid,
    kmp_i18n_msg_AffIgnoringNotAvailable,
    kmp_i18n_msg_last,

    // Set #5, hints.
    kmp_i18n_hnt_first = 327680,
    kmp_i18n_hnt_SubmitBugReport,
    kmp_i18n_hnt_OBSOLETE46,
    kmp_i18n_hnt_ChangeStackLimit,
    kmp_i18n_hnt_Unset_ALL_THREADS,
    kmp_i18n_hnt_Set_ALL_THREADPRIVATE,
    kmp_i18n_hnt_PossibleSystemLimitOnThreads,
    kmp_i18n_hnt_DuplicateLibrary,
    kmp_i18n_hnt_NameComesFrom_CPUINFO_FILE,
    kmp_i18n_hnt_NotEnoughMemory,
    kmp_i18n_hnt_ValidBoolValues,
    kmp_i18n_hnt_BufferOverflow,
    kmp_i18n_hnt_RunningAtMaxPriority,
    kmp_i18n_hnt_ChangeMonitorStackSize,
    kmp_i18n_hnt_ChangeWorkerStackSize,
    kmp_i18n_hnt_IncreaseWorkerStackSize,
    kmp_i18n_hnt_DecreaseWorkerStackSize,
    kmp_i18n_hnt_Decrease_NUM_THREADS,
    kmp_i18n_hnt_IncreaseMonitorStackSize,
    kmp_i18n_hnt_DecreaseMonitorStackSize,
    kmp_i18n_hnt_DecreaseNumberOfThreadsInUse,
    kmp_i18n_hnt_DefaultScheduleKindUsed,
    kmp_i18n_hnt_GetNewerLibrary,
    kmp_i18n_hnt_CheckEnvVar,
    kmp_i18n_hnt_OBSOLETE47,
    kmp_i18n_hnt_OBSOLETE48,
    kmp_i18n_hnt_BadExeFormat,
    kmp_i18n_hnt_SystemLimitOnThreads,
    kmp_i18n_hnt_SetNewBound,
    kmp_i18n_hnt_ValidValuesRange,
    kmp_i18n_hnt_last,

    kmp_i18n_xxx_lastest

}; // enum kmp_i18n_id

typedef enum kmp_i18n_id  kmp_i18n_id_t;


// end of file //
